(**********************************************************************
 * info-401, programmation fonctionnelle
 * fichier pour le TP5
 * Nom : Luke Skywalker
 *********************************************************************)


(*
 * difficulty of the game...
 *)
type level = {
  mutable new_proba : float; (* mean number of number of sphere per second *)
  mutable max_speed : float; (* maximum speed of spheres *)
}

(* initial difficulty *)
let level = { new_proba = 1.0;  max_speed = 1.0; }

(* increase difficulty *)
let increase_level refresh =
  level.new_proba <- level.new_proba +. 1.0 /. refresh /. 60.0;
  level.max_speed <- level.max_speed +. 1.0 /. refresh /. 60.0

exception End_of_game


(*
 * type to store ship position
 *)
type ship = {
  mutable our_pos : float * float * float;
  mutable our_x_speed : float;
  mutable our_y_speed : float;
  mutable our_points : int
}

(* initial position *)
let ship = {
  our_pos = (0.0, 0.0, 0.0);
  our_x_speed = 0.0;
  our_y_speed = 0.0;
  our_points = 0;
}


(*
 * spheres...
 *)
type 'a sphere = {
  mutable sphere_pos : float * float * float;       (* position *)
  mutable sphere_speed : float * float * float;     (* speed vector *)
  sphere_radius : float;                            (* radius *)
  sphere_color : float * float * float * float;     (* color *)
  sphere_user_data : 'a;                            (* additional data *)
}

let first_sphere a =
  {
    sphere_pos = (0.0, 0.0, -10.0);
    sphere_radius = 0.5;
    sphere_speed = (0.0, 0.0, 1.0);
    sphere_color = Random.color ();
    sphere_user_data = a;
   }

let random_sphere a =
  {
    sphere_pos = (Random.float 2.0 -. 1.0, Random.float 2.0 -. 1.0, -10.0);
    sphere_speed =
      (Random.float 0.2 -. 0.1, Random.float 0.2 -. 0.1, Random.float 1.0 *. level.max_speed);
    sphere_radius = Random.float 1.0;
    sphere_color = Random.color ();
    sphere_user_data = a;
  }

(*
 * functions on spheres...
 *)

(* test is two spheres intersect and have speed vector that rapproches them.
   it yes, change speed arcording to a physical model where all sphere have
   same density *)
let bounce s1 s2 =
  let sep = Vector.sub s2.sphere_pos s1.sphere_pos in
  if s1 != s2 && Vector.norm sep < s1.sphere_radius +. s2.sphere_radius then begin
    let normal = Vector.normalize sep in
    let d1 = Vector.dot normal s1.sphere_speed in
    let d2 = Vector.dot normal s2.sphere_speed in
    let d = d1 -. d2 in
    if d > 0.0 then begin
      let m1 =  s1.sphere_radius *. s1.sphere_radius *. s1.sphere_radius in
      let m2 =  s2.sphere_radius *. s2.sphere_radius *. s2.sphere_radius in
      let alpha1 = -2.0 *. d /. (1.0 +. m1 /. m2) in
      let alpha2 = 2.0 *. d /. (1.0 +. m2 /. m1) in
      s1.sphere_speed <- Vector.add_alpha s1.sphere_speed alpha1 normal;
      s2.sphere_speed <- Vector.add_alpha s2.sphere_speed alpha2 normal;
    end
  end


(****************************************************************************
 ****************************************************************************
 *** Les fonctions  crire commencent ici... *******************************
 ****************************************************************************
 ***************************************************************************)

(* fonction pour tester si deux spheres s'intersectent *)
let intersect s1 s2 : bool = false  (*TODO*)


(* dplace la sphere (1er argument) le long de l'axe des "z" pour viter qu'elle
 * intersecte une des autres spheres de la liste (second argument)
 * Cette fonction est uniquement appele sur les nouvelles spheres... *)
let avoid_others (s:'a sphere) (spheres:'a sphere list) : unit = () (*TODO*)


(* vrifie les collisions, et mets les vecteurs vitesse  jours en utilisant la
 * fonction "bounce" dfinie plus haut *)
let check_collisions_between (spheres:'a sphere list) : unit = () (*TODO*)


(* fonction de mise  jour appele avant les affichages ("refresh" fois par
 * seconde...).
 * Il faut :
 *   - vrifier si on n'est pas rentr dans une sphere (et arreter la partie)
 *   - supprimer les spheres qu'on a dpasses (et augmenter le score d'un
 *     point et supprimer la sphere de la liste)
 *   - supprimer de la liste les spheres qui sont trop loin
 *   - mettre  jour les positions (en fonction du vecteur vitesse) des spheres
 *     restantes
 *)
let move_spheres_and_check refresh ship (spheres:'a sphere list) : 'a sphere list =
  List.filter  (* on ne va garder que certaines spheres *)
    (fun s ->
      let (x,y,z) = s.sphere_pos in
      if z >= 0.
      (* si la sphere est derriere nous (z>0), on la supprime *)
      then false
      else
      (* sinon, on va changer sa position en fonction de sa vitesse,
       * on la garde *)
        begin
          (* TODO : mettre la position  jours *)
          true
        end) spheres

